@@ -1,5 +1,5 @@
 #
-# Copyright (c), 2018-2020, SISSA (International School for Advanced Studies).
+# Copyright (c), 2018-2021, SISSA (International School for Advanced Studies).
 # All rights reserved.
 # This file is distributed under the terms of the MIT License.
 # See the file 'LICENSE' in the root directory of the present
@@ -16,7 +16,8 @@ In XPath there are 7 kinds of nodes:
     element, attribute, text, namespace, processing-instruction, comment, document
 
 Element-like objects are used for representing elements and comments, ElementTree-like objects
-for documents. Generic tuples are used for representing attributes and named-tuples for namespaces.
+for documents.
+XPathNode subclasses are used for representing other node types and typed elements/attributes.
 """
 import locale
 import contextlib
@@ -24,15 +25,17 @@ import math
 from copy import copy
 from decimal import Decimal
 from itertools import product
+from typing import Union
 import urllib.parse
+from xml.etree.ElementTree import Element
 
 from .exceptions import ElementPathError, ElementPathValueError, XPATH_ERROR_CODES
-from .namespaces import XQT_ERRORS_NAMESPACE, XSD_NAMESPACE, XPATH_FUNCTIONS_NAMESPACE, \
+from .namespaces import XQT_ERRORS_NAMESPACE, XSD_NAMESPACE, \
     XSD_ANY_TYPE, XSD_ANY_SIMPLE_TYPE, XSD_ANY_ATOMIC_TYPE
-from .xpath_nodes import AttributeNode, TextNode, NamespaceNode, TypedAttribute, \
-    TypedElement, is_etree_element, etree_iter_strings, is_comment_node, \
-    is_processing_instruction_node, is_element_node, is_document_node, \
-    is_xpath_node, is_schema_node
+from .xpath_nodes import XPathNode, TypedElement, AttributeNode, TextNode, \
+    NamespaceNode, TypedAttribute, is_etree_element, etree_iter_strings, \
+    is_comment_node, is_processing_instruction_node, is_element_node, \
+    is_document_node, is_xpath_node, is_schema_node
 from .datatypes import xsd10_atomic_types, xsd11_atomic_types, AbstractDateTime, \
     AnyURI, UntypedAtomic, Timezone, DateTime10, Date10, DayTimeDuration, Duration, \
     Integer, DoubleProxy10, DoubleProxy, QName
@@ -375,17 +378,19 @@ class XPathToken(Token):
         self.parser.check_variables(context.variables)
 
         for result in self.select(context):
-            if isinstance(result, (TypedElement, TextNode)):
-                yield result[0]
-            elif isinstance(result, AttributeNode):
-                yield result[1]
-            elif isinstance(result, TypedAttribute):
-                if is_schema_node(result[0][1]):
-                    yield result[0][1]
-                else:
-                    yield result[-1]
-            else:
+            if not isinstance(result, XPathNode):
                 yield result
+            elif isinstance(result, (TextNode, AttributeNode)):
+                yield result.value
+            elif isinstance(result, TypedElement):
+                yield result.elem
+            elif isinstance(result, TypedAttribute):
+                if is_schema_node(result.attribute.value):
+                    yield result.attribute.value
+                else:
+                    yield result.value
+            elif isinstance(result, NamespaceNode):  # pragma: no cover
+                yield result.prefix, result.uri
 
     def get_results(self, context):
         """
@@ -395,18 +400,12 @@ class XPathToken(Token):
         :return: a list or a simple datatype when the result is a single simple type \
         generated by a literal or function token.
         """
-        self.parser.check_variables(context.variables)
-
         results = [x for x in self.select_results(context)]
         if len(results) == 1:
             res = results[0]
             if isinstance(res, (bool, int, float, Decimal)):
                 return res
-            elif isinstance(res, tuple) or is_etree_element(res) or is_document_node(res):
-                return results
-            elif is_schema_node(res):
-                return results
-            elif self.symbol in ('text', 'node'):
+            elif is_etree_element(res) or is_document_node(res) or is_schema_node(res):
                 return results
             elif self.label in ('function', 'literal'):
                 return res
@@ -512,7 +511,7 @@ class XPathToken(Token):
         """
         if self.symbol in ('(name)', '*'):
             pass
-        elif namespace == XPATH_FUNCTIONS_NAMESPACE:
+        elif namespace == self.parser.function_namespace:
             if self.label != 'function':
                 msg = "a name, a wildcard or a function expected"
                 raise self.wrong_syntax(msg, code='XPST0017')
@@ -606,11 +605,11 @@ class XPathToken(Token):
             if xsd_node is None:
                 if name == schema_context.root.tag == '{%s}schema' % XSD_NAMESPACE:
                     yield None
-                continue
+                continue  # pragma: no cover
 
             try:
                 if isinstance(xsd_node, AttributeNode):
-                    if xsd_node[1].is_matching(name):
+                    if xsd_node.value.is_matching(name):
                         if xsd_node.name is None:
                             # node is an XSD attribute wildcard
                             xsd_node = self.parser.schema.get_attribute(name)
@@ -645,9 +644,9 @@ class XPathToken(Token):
         if isinstance(item, AttributeNode):
             item = item.value
         elif isinstance(item, TypedAttribute):
-            item = item[0].value
+            item = item.attribute.value
         elif isinstance(item, TypedElement):
-            item = item[0]
+            item = item.elem
 
         if not is_schema_node(item):
             return
@@ -678,9 +677,9 @@ class XPathToken(Token):
         elif isinstance(item, str):
             xsd_type = self.xsd_types.get(item)
         elif isinstance(item, AttributeNode):
-            xsd_type = self.xsd_types.get(item[0])
+            xsd_type = self.xsd_types.get(item.name)
         elif isinstance(item, (TypedAttribute, TypedElement)):
-            return item.type
+            return item.xsd_type
         else:
             xsd_type = self.xsd_types.get(item.tag)
 
@@ -690,9 +689,9 @@ class XPathToken(Token):
             return xsd_type
         elif isinstance(item, AttributeNode):
             for x in xsd_type:
-                if x.is_valid(item[1]):
+                if x.is_valid(item.value):
                     return x
-        elif not isinstance(item, str):
+        elif is_etree_element(item):
             for x in xsd_type:
                 if x.is_simple():
                     if x.is_valid(item.text):
@@ -702,7 +701,7 @@ class XPathToken(Token):
 
         return xsd_type[0]
 
-    def get_typed_node(self, item):
+    def get_typed_node(self, item: Union[Element, AttributeNode]):
         """
         Returns a typed node if the item is matching an XSD type.
 
@@ -712,8 +711,8 @@ class XPathToken(Token):
           https://www.w3.org/TR/xquery-semantics/
 
         :param item: an untyped attribute ot element.
-        :return: a TypedAttribute or a TypedElement, or the argument \
-        if it's not matching any associated XSD type.
+        :return: a typed AttributeNode/ElementNode if the argument is matching \
+        any associated XSD type.
         """
         if isinstance(item, (TypedAttribute, TypedElement)):
             return item
@@ -723,7 +722,7 @@ class XPathToken(Token):
             return item
         elif xsd_type.name in XSD_SPECIAL_TYPES:
             if isinstance(item, AttributeNode):
-                return TypedAttribute(item, xsd_type, UntypedAtomic(item[1]))
+                return TypedAttribute(item, xsd_type, UntypedAtomic(item.value))
             return TypedElement(item, xsd_type, UntypedAtomic(item.text or ''))
 
         elif xsd_type.has_mixed_content():
@@ -751,7 +750,7 @@ class XPathToken(Token):
 
             try:
                 if isinstance(item, AttributeNode):
-                    return TypedAttribute(item, xsd_type, builder(item[1]))
+                    return TypedAttribute(item, xsd_type, builder(item.value))
                 else:
                     return TypedElement(item, xsd_type, builder(item.text))
             except (TypeError, ValueError):
@@ -771,8 +770,8 @@ class XPathToken(Token):
 
         try:
             if isinstance(item, AttributeNode):
-                if xsd_type.is_valid(item[1]):
-                    return TypedAttribute(item, xsd_type, builder(item[1]))
+                if xsd_type.is_valid(item.value):
+                    return TypedAttribute(item, xsd_type, builder(item.value))
             elif xsd_type.is_valid(item.text):
                 return TypedElement(item, xsd_type, builder(item.text))
         except (TypeError, ValueError):
@@ -840,14 +839,14 @@ class XPathToken(Token):
         """
         if obj is None:
             return
-        elif isinstance(obj, tuple):
-            if isinstance(obj, (TypedElement, TypedAttribute)):
-                return obj[-1]
-            elif isinstance(obj, NamespaceNode):
-                return obj[1]
-            return UntypedAtomic(obj[-1])
+        elif isinstance(obj, XPathNode):
+            if isinstance(obj, (AttributeNode, TextNode)):
+                return UntypedAtomic(obj.value)
+            return obj.value  # a typed node or a NamespaceNode
+
         elif is_schema_node(obj):
             return self.parser.get_atomic_value(obj.type)
+
         elif hasattr(obj, 'tag'):
             if is_comment_node(obj):
                 return obj.text
@@ -855,6 +854,7 @@ class XPathToken(Token):
                 return obj.text
             elif hasattr(obj, 'attrib') and hasattr(obj, 'text'):
                 return UntypedAtomic(''.join(etree_iter_strings(obj)))
+
         elif is_document_node(obj):
             value = ''.join(etree_iter_strings(obj.getroot()))
             return UntypedAtomic(value)
@@ -867,15 +867,15 @@ class XPathToken(Token):
         """
         if obj is None:
             return ''
-        elif isinstance(obj, tuple):
+        elif isinstance(obj, XPathNode):
             if isinstance(obj, TypedElement):
-                if obj[-1] is None:
+                if obj.value is None:
                     return ''.join(etree_iter_strings(obj))
-                return str(obj[-1])
+                return str(obj.value)
             elif isinstance(obj, (AttributeNode, TypedAttribute)):
-                return str(obj[-1])
-            elif isinstance(obj, (TextNode, NamespaceNode)):
-                return obj[-1]
+                return str(obj.value)
+            else:
+                return obj.value  # TextNode or NamespaceNode
         elif is_schema_node(obj):
             return str(self.parser.get_atomic_value(obj.type))
         elif hasattr(obj, 'tag'):
@@ -943,7 +943,7 @@ class XPathToken(Token):
         :param message_or_error: an optional custom additional message.
         """
         if isinstance(code, QName):
-            code, namespace = code.local_name, code.namespace
+            code, namespace = code.local_name, code.uri
         elif ':' not in code:
             namespace = None
         else:
